/**
 * @file presence.c Galago Presence message processor
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "filter.h"
#include "meta-presence.h"
#include "utils.h"

static void
_galagod_dbus_presence_emit_status_added(GalagoPresence *presence,
										 GalagoStatus *status)
{
	DBusMessage *message;
	DBusMessageIter iter;

	galago_return_if_fail(presence != NULL);
	galago_return_if_fail(status   != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(presence),
									  GALAGO_DBUS_PRESENCE_INTERFACE,
									  "StatusAdded");

	dbus_message_iter_init(message, &iter);
	galago_dbus_message_iter_append_object(&iter, status);

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static void
_galagod_dbus_presence_emit_status_removed(GalagoPresence *presence,
										   const char *status_id)
{
	DBusMessage *message;
	DBusMessageIter iter;

	galago_return_if_fail(presence  != NULL);
	galago_return_if_fail(status_id != NULL && *status_id != '\0');

	message = dbus_message_new_signal(galago_object_get_dbus_path(presence),
									  GALAGO_DBUS_PRESENCE_INTERFACE,
									  "StatusRemoved");

	dbus_message_iter_init(message, &iter);
	dbus_message_iter_append_string(&iter, status_id);

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static DBusHandlerResult
presence_add_status(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPresence *presence)
{
	GalagoStatus *status;
	GalagodMetaPresence *meta_presence;
	GalagoPresence *main_presence;
	DBusMessageIter iter;
	const GalagoList* l;
	const char *status_id;
	int status_weight = 0;

	meta_presence = GALAGOD_META_PRESENCE(presence);

	dbus_message_iter_init(message, &iter);
	status = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_STATUS);
	dbus_message_iter_next(&iter);

	status_id = galago_status_get_id(status);

	if (!galago_presence_has_status(presence, status_id))
		galago_presence_add_status(presence, galago_status_duplicate(status));

	for (l = galagod_meta_presence_get_presences(meta_presence);
		 l != NULL;
		 l = l->next)
	{
		GalagoPresence *temp_presence = (GalagoPresence *)l->data;

		if (galago_presence_has_status(temp_presence, status_id))
			status_weight++;
		else
			status_weight--;
	}

	main_presence = galagod_meta_presence_get_presence(meta_presence);

	if (status_weight > 0)
	{
		if (!galago_presence_has_status(main_presence, status_id))
		{
			galago_context_push(galagod_get_context());

			galago_presence_add_status(main_presence,
									   galago_status_duplicate(status));

			galago_context_pop();

			_galagod_dbus_presence_emit_status_added(main_presence, status);
		}
	}

	galago_object_unref(status);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
presence_remove_status(DBusConnection *dbus_conn, DBusMessage *message,
					   GalagodCoCo *coco, GalagoPresence *presence)
{
	char *status_id;
	GalagodMetaPresence *meta_presence;
	GalagoPresence *main_presence;
	DBusMessageIter iter;

	meta_presence = GALAGOD_META_PRESENCE(presence);

	dbus_message_iter_init(message, &iter);
	status_id = dbus_message_iter_get_string(&iter);

	galago_presence_remove_status(presence, status_id);

	main_presence = galagod_meta_presence_get_presence(meta_presence);

	if (galago_presence_has_status(main_presence, status_id))
	{
		galago_presence_remove_status(main_presence, status_id);

		_galagod_dbus_presence_emit_status_removed(main_presence, status_id);
	}

	dbus_free(status_id);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
presence_set_statuses(DBusConnection *dbus_conn, DBusMessage *message,
					  GalagodCoCo *coco, GalagoPresence *presence)
{
	GalagodMetaPresence *meta_presence;
	GalagoList *list, *list2 = NULL, *l;
	DBusMessageIter iter;

	meta_presence = GALAGOD_META_PRESENCE(presence);

	dbus_message_iter_init(message, &iter);
	list = galago_dbus_message_iter_get_object_list(&iter,
													GALAGO_CLASS_STATUS);

	galago_presence_set_statuses(presence, list);

	galago_context_push(galagod_get_context());

	for (l = list; l != NULL; l = l->next)
	{
		list2 = galago_list_append(list2,
			galago_status_duplicate((GalagoStatus *)l->data));
	}

	galago_presence_set_statuses(
		galagod_meta_presence_get_presence(meta_presence), list2);

	galago_context_pop();

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
presence_set_idle(DBusConnection *dbus_conn, DBusMessage *message,
				  GalagodCoCo *coco, GalagoPresence *presence)
{
	GalagodMetaPresence *meta_presence;
	GalagoPresence *main_presence;
	DBusMessageIter iter;
	galago_bool idle;
	dbus_uint32_t idle_time = 0;

	meta_presence = GALAGOD_META_PRESENCE(presence);

	dbus_message_iter_init(message, &iter);
	idle = dbus_message_iter_get_boolean(&iter);
	dbus_message_iter_next(&iter);

	if (idle)
		idle_time = dbus_message_iter_get_uint32(&iter);

	main_presence = galagod_meta_presence_get_presence(meta_presence);

	galago_presence_set_idle(presence,      idle, idle_time);
	galago_presence_set_idle(main_presence, idle, idle_time);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
presence_path_message_func(DBusConnection *dbus_conn, DBusMessage *message,
						   void *user_data)
{
	DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	GalagodStateData *data = (GalagodStateData *)user_data;

	if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
		return result;

	galago_log_misc("%s: %s\n",
					dbus_message_get_interface(message),
					dbus_message_get_member(message));

	galago_context_push(data->context);

	if (dbus_message_is_method_call(message,
									GALAGO_DBUS_PRESENCE_INTERFACE,
									"AddStatus"))
	{
		result = presence_add_status(dbus_conn, message, data->coco,
									 data->u.presence);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PRESENCE_INTERFACE,
										 "RemoveStatus"))
	{
		result = presence_remove_status(dbus_conn, message, data->coco,
										data->u.presence);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PRESENCE_INTERFACE,
										 "SetStatuses"))
	{
		result = presence_set_statuses(dbus_conn, message, data->coco,
									   data->u.presence);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PRESENCE_INTERFACE,
										 "SetIdle"))
	{
		result = presence_set_idle(dbus_conn, message, data->coco,
								   data->u.presence);
	}
	else
	{
		galago_log_warning("Unhandled message!\n");
		galago_log_misc("\tobj_path = '%s'\n",
						dbus_message_get_path(message));
		galago_log_misc("\tinterface = '%s'\n",
						dbus_message_get_interface(message));
		galago_log_misc("\tmethod = '%s'\n",
						dbus_message_get_member(message));
	}

	galago_context_pop();

	return result;
}

void
galagod_presence_add_filter(GalagoPresence *presence)
{
	galagod_filters_add(presence, presence_path_message_func);
}

void
galagod_presence_remove_filter(GalagoPresence *presence)
{
	galagod_filters_remove(presence);
}
